# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any, List, Optional

import numpy as np

from audiolab.av import AudioGraph, aformat, load_url, split_audio_frame
from audiolab.av.typing import AudioFormat, Dtype, Filter, Seconds
from audiolab.reader.info import Info


class Reader(Info):
    def __init__(
        self,
        file: Any,
        stream_id: int = 0,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        filters: Optional[List[Filter]] = None,
        dtype: Optional[Dtype] = None,
        is_planar: bool = False,
        format: Optional[AudioFormat] = None,
        rate: Optional[int] = None,
        to_mono: bool = False,
        frame_size: Optional[int] = None,
        frame_size_ms: Optional[int] = None,
        return_ndarray: bool = True,
        cache_url: bool = True,
    ):
        """
        Create a Reader object.

        Args:
            file: The audio file, path to audio file, bytes of audio data, etc.
            stream_id: The index of the stream to load.
            offset: The offset of the audio stream to load.
            duration: The duration of the audio stream to load.
            filters: The filters to apply to the audio stream.
            dtype: The data type of the audio frames.
            is_planar: Whether the audio frames are planar.
            format: The format of the audio frames.
            rate: The sample rate of the audio frames.
            to_mono: Whether to convert the audio frames to mono.
            frame_size: The frame size of the audio frames.
            frame_size_ms: The frame size in milliseconds of the audio frames.
            return_ndarray: Whether to return the audio frames as ndarrays.
            cache_url: Whether to cache the audio file.
        Returns:
            The Reader object.
        """
        if isinstance(file, str) and "://" in file and cache_url:
            file = load_url(file)

        super().__init__(file, stream_id)
        self.start_time = int(offset / self.stream.time_base)
        self.end_time = Seconds("inf") if duration is None else offset + duration
        if self.start_time > 0:
            self.container.seek(self.start_time, any_frame=True, stream=self.stream)

        if not all([dtype is None, format is None, rate is None, not to_mono]):
            filters = filters or []
            filters.append(aformat(dtype, is_planar, format, rate, to_mono))

        if frame_size_ms is not None:
            frame_size = frame_size_ms * self.stream.rate / 1000
        else:
            frame_size = frame_size or np.iinfo(np.uint32).max
        self.frame_size = min(frame_size, np.iinfo(np.uint32).max)
        self.graph = AudioGraph(
            stream=self.stream, filters=filters, frame_size=frame_size, return_ndarray=return_ndarray
        )

    @property
    def num_frames(self) -> int:
        """
        Get the number of the input audio frames in the audio stream.

        Returns:
            The number of the input audio frames in the audio stream.
        """
        return math.ceil(self.duration * self.rate / self.frame_size)

    def __iter__(self):
        for frame in self.container.decode(self.stream):
            assert frame.time == float(frame.pts * self.stream.time_base)
            if frame.time > self.end_time:
                break
            if frame.time + frame.samples / frame.rate > self.end_time:
                frame, _ = split_audio_frame(frame, self.end_time - frame.time)
            self.graph.push(frame)
            yield from self.graph.pull()
        yield from self.graph.pull(partial=True)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from io import BytesIO
from typing import Any, Dict, Optional

import av
import numpy as np

from audiolab.av import from_ndarray
from audiolab.av.format import get_format
from audiolab.av.typing import AudioFormat, AudioFrame, AudioLayout, Codec, ContainerFormat, Dtype

logger = logging.getLogger(__name__)


class Writer:
    def __init__(
        self,
        file: Any,
        rate: int,
        codec: Optional[Codec] = None,
        channels: Optional[int] = None,
        dtype: Optional[Dtype] = None,
        is_planar: Optional[bool] = None,
        format: Optional[AudioFormat] = None,
        layout: Optional[AudioLayout] = None,
        container_format: Optional[ContainerFormat] = None,
        options: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        """
        Create a Writer object.

        Args:
            file: The audio file, path to audio file, bytes, etc.
            rate: The sample rate of the audio stream.
            codec: The codec of the audio container.
            channels: The number of channels of the audio stream.
            dtype: The data type of the audio stream.
            is_planar: Whether the audio stream is planar.
            format: The format of the audio stream.
            layout: The layout of the audio stream.
            container_format: The format of the audio container.
            options: The options of the audio stream.
        Returns:
            The Writer object.
        """
        if isinstance(file, BytesIO):
            assert container_format is not None
            if isinstance(container_format, av.ContainerFormat):
                container_format = container_format.name
            self.container = av.open(file, "w", container_format)
        else:
            self.container = av.open(file, "w")
        # set and check codec
        codec = codec or self.container.default_audio_codec
        if isinstance(codec, str):
            codec = av.Codec(codec, "w")
        assert codec.name in self.container.supported_codecs
        # set and check format
        if dtype is not None:
            format = get_format(dtype, is_planar, codec.audio_formats)
        else:
            format = format or codec.audio_formats[0]
        if isinstance(format, av.AudioFormat):
            format = format.name
        assert format in set(format.name for format in codec.audio_formats)
        if layout is None:
            assert channels in (1, 2)
            layout = "mono" if channels == 1 else "stereo"

        kwargs = {**kwargs, **{"format": format, "layout": layout}}
        self.stream = self.container.add_stream(codec.name, rate, options, **kwargs)

    def write(self, frame: AudioFrame):
        """
        Write an audio frame to the audio stream.

        Args:
            frame: The audio frame to write.
        """
        if isinstance(frame, np.ndarray):
            frame = np.atleast_2d(frame)
            assert frame.ndim == 2, "Audio frame must be 1D (samples,) or 2D (channels, samples)"
            assert frame.shape[0] == self.stream.channels, "Number of channels in frame does not match stream"
            frame = from_ndarray(frame, self.stream.format.name, self.stream.layout, self.stream.rate)
        for packet in self.stream.encode(frame):
            self.container.mux(packet)

    def close(self):
        """
        Close the audio stream.
        """
        self.container.close()



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib.resources import files
from io import BytesIO

from av import AudioCodecContext
from jinja2 import Environment, FileSystemLoader

loader = FileSystemLoader(files("audiolab.av").joinpath("templates"))


def get_template(name: str) -> str:
    """
    Get a template from the templates directory.

    Args:
        name: The name of the template.
    Returns:
        The template.
    """
    return Environment(loader=loader).get_template(f"{name}.txt")


def load_url(url: str) -> BytesIO:
    """
    Load an audio file from a URL.

    Args:
        url: The URL of the audio file.
    Returns:
        The audio bytes.
    """
    from lhotse.caching import AudioCache
    from lhotse.utils import SmartOpen

    audio_bytes = AudioCache.try_cache(url)
    if audio_bytes is None:
        with SmartOpen.open(url, "rb") as f:
            audio_bytes = f.read()
        AudioCache.add_to_cache(url, audio_bytes)
    return BytesIO(audio_bytes)


def is_streamable(codec_context: AudioCodecContext) -> bool:
    """
    Check if the codec is streamable.

    Args:
        codec_context: The codec context.
    Returns:
        Whether the codec is streamable.

    Note:
        * https://github.com/FFmpeg/FFmpeg/blob/master/libavcodec/avcodec.h#L1045-L1051
        * Each submitted frame except the last must contain exactly frame_size samples per channel.
        * May be 0 when the codec has AV_CODEC_CAP_VARIABLE_FRAME_SIZE set, then the frame size is not restricted.
    """
    return codec_context.frame_size in (0, 1)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Union

import av
from av import AudioLayout, AudioStream, Codec, time_base
from av.container import Container
from humanize import naturalsize

from audiolab.av.typing import Seconds
from audiolab.av.utils import get_template, is_streamable


class Info:
    stream: AudioStream
    channels: int
    codec: Codec
    rate: int
    sample_rate: int
    layout: AudioLayout

    def __init__(self, file: Any, stream_id: int = 0, force_decoding: bool = False):
        """
        Create an Info object.

        Args:
            file: The input audio file, path to audio file, bytes of audio data, etc.
            stream_id: The index of the stream to get information from.
            force_decoding: Whether to force decoding the audio file to get the duration.
        """
        self.container = av.open(file)
        self.stream = self.container.streams.audio[stream_id]
        self.channels = self.stream.channels
        self.codec = self.stream.codec
        self.rate = self.stream.rate
        self.sample_rate = self.stream.sample_rate
        self.layout = self.stream.layout
        self.precision = self.stream.format.bits
        self.metadata = {**self.container.metadata, **self.stream.metadata}
        self.is_streamable = is_streamable(self.stream.codec_context)

        # number of audio samples (per channel)
        self.num_samples = None
        self.duration = None
        if self.stream.duration:
            start_time = self.stream.start_time or 0
            self.duration = Seconds((self.stream.duration + start_time) * self.stream.time_base)
            self.num_samples = int(self.duration * self.stream.rate)
        elif self.container.duration:
            start_time = self.container.start_time or 0
            self.duration = Seconds((self.container.duration + start_time) / time_base)
            self.num_samples = int(self.duration * self.stream.rate)
        elif force_decoding:
            # decode the stream to get the duration if the duration is not available
            self.num_samples = 0
            for frame in self.container.decode(self.stream):
                self.num_samples += frame.samples
            self.duration = Seconds(self.num_samples / self.stream.rate)

    @property
    def bit_rate(self) -> Union[int, None]:
        """
        Get the bit rate of the audio stream.

        Returns:
            The bit rate of the audio stream.
        """
        bit_rate = self.stream.bit_rate or self.container.bit_rate
        if bit_rate in (None, 0) and self.duration is not None:
            # bytes * 8 / seconds
            bit_rate = self.container.size * 8 / self.duration
        return bit_rate

    @property
    def num_cdda_sectors(self) -> Union[float, None]:
        """
        Get the number of CDDA sectors of the audio stream.

        Returns:
            The number of CDDA sectors of the audio stream.
        """
        return None if self.duration is None else round(self.duration * 75, 2)

    @staticmethod
    def rstrip_zeros(s: Union[int, float, str]) -> str:
        """
        Remove trailing zeros from a string.

        Args:
            s: The string to remove trailing zeros from.
        Returns:
            The string with trailing zeros removed.
        """
        if not isinstance(s, str):
            s = str(s)
        return " ".join(x.rstrip("0").rstrip(".") for x in s.split())

    @staticmethod
    def format_bit_rate(bit_rate: Union[int, None]) -> str:
        """
        Format the bit rate of the audio stream.

        Args:
            bit_rate: The bit rate of the audio stream.
        Returns:
            The formatted bit rate of the audio stream.
        """
        if bit_rate is None or bit_rate <= 0:
            return "N/A"
        bit_rate = naturalsize(bit_rate).rstrip("B")
        return Info.rstrip_zeros(bit_rate) + "bps"

    @staticmethod
    def format_duration(duration: Union[Seconds, None]) -> str:
        """
        Format the duration of the audio stream.

        Args:
            duration: The duration of the audio stream.
        Returns:
            The formatted duration of the audio stream.
        """
        if duration is None:
            return "N/A"
        hours, rest = divmod(duration, 3600)
        minutes, seconds = divmod(rest, 60)
        return f"{int(hours):02d}:{int(minutes):02d}:{seconds:06.3f}"

    @staticmethod
    def format_name(container: Container) -> str:
        """
        Format the name of the container.

        Args:
            container: The container.
        Returns:
            The formatted name of the container.
        """
        name = container.name
        _format_name = container.format.name
        if name.lower().endswith(_format_name.lower()):
            return f"'{name}'"
        # <none> for BytesIO, <stdin> for stdin
        if name in ("<none>", "<stdin>"):
            return f"{name} ({_format_name})"
        return f"'{name}' ({_format_name})"

    @staticmethod
    def format_num_cdda_sectors(num_cdda_sectors: Union[float, None]) -> str:
        """
        Format the number of CDDA sectors of the audio stream.

        Args:
            num_cdda_sectors: The number of CDDA sectors of the audio stream.
        Returns:
            The formatted number of CDDA sectors of the audio stream.
        """
        return "N/A" if num_cdda_sectors is None else Info.rstrip_zeros(num_cdda_sectors)

    @staticmethod
    def format_size(size: int) -> str:
        """
        Format the size of the audio file.

        Args:
            size: The size of the audio file.
        Returns:
            The formatted size of the audio file.
        """
        if size in (-1, -78):
            return "N/A"
        size = naturalsize(size)
        return Info.rstrip_zeros(size)

    def __str__(self):
        """
        Format the audio file information.

        Returns:
            The formatted audio file information.
        """
        return get_template("info").render(
            name=Info.format_name(self.container),
            channels=self.channels,
            rate=self.rate,
            precision=self.precision,
            duration=Info.format_duration(self.duration),
            num_samples=self.num_samples or "N/A",
            num_cdda_sectors=Info.format_num_cdda_sectors(self.num_cdda_sectors),
            size=Info.format_size(self.container.size),
            bit_rate=Info.format_bit_rate(self.bit_rate),
            codec=self.codec.long_name,
            metadata=self.metadata,
        )



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Iterator, List, Optional, Tuple, Union

import numpy as np

from audiolab.av.typing import AudioFormat, Dtype, Filter, Seconds
from audiolab.reader.info import Info
from audiolab.reader.reader import Reader
from audiolab.reader.stream_reader import StreamReader


def info(file: Any, stream_id: int = 0, force_decode: bool = False) -> Info:
    """
    Get the information of an audio file.

    Args:
        file: The input audio file, path to audio file, bytes of audio data, etc.
        stream_id: The index of the stream to get information from.
        force_decode: Whether to force decoding the audio file to get the duration.
    Returns:
        The information of the audio file.
    """
    return Info(file, stream_id, force_decode)


def load_audio(
    file: Any,
    stream_id: int = 0,
    offset: Seconds = 0.0,
    duration: Optional[Seconds] = None,
    filters: Optional[List[Filter]] = None,
    dtype: Optional[Dtype] = None,
    is_planar: bool = False,
    format: Optional[AudioFormat] = None,
    rate: Optional[int] = None,
    to_mono: bool = False,
    frame_size: Optional[int] = None,
    frame_size_ms: Optional[int] = None,
    return_ndarray: bool = True,
    cache_url: bool = True,
) -> Union[Iterator[Tuple[np.ndarray, int]], Tuple[np.ndarray, int]]:
    """
    Load an audio file.

    Args:
        file: The audio file, path to audio file, bytes of audio data, etc.
        stream_id: The index of the stream to load.
        offset: The offset of the audio stream to load.
        duration: The duration of the audio stream to load.
        filters: The filters to apply to the audio stream.
        dtype: The data type of the audio frames.
        is_planar: Whether the audio frames are planar.
        format: The format of the audio frames.
        rate: The sample rate of the audio frames.
        to_mono: Whether to convert the audio frames to mono.
        frame_size: The frame size of the audio frames.
        frame_size_ms: The frame size in milliseconds of the audio frames.
        return_ndarray: Whether to return the audio frames as ndarrays.
        cache_url: Whether to cache the audio file.
    Returns:
        The audio frames and the sample rate of the audio stream.
    """
    reader = Reader(
        file,
        stream_id,
        offset,
        duration,
        filters,
        dtype,
        is_planar,
        format,
        rate,
        to_mono,
        frame_size,
        frame_size_ms,
        return_ndarray,
        cache_url,
    )
    generator = reader.__iter__()
    if reader.frame_size < np.iinfo(np.uint32).max:
        return generator
    return next(generator)


__all__ = ["Reader", "StreamReader", "load_audio"]



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import click

import audiolab
from audiolab.reader.info import Info


@click.command()
@click.argument("audio-files", type=click.File(mode="rb"), nargs=-1)
@click.option("-S", "--stream-id", type=int, default=0)
@click.option("-f", "--force-decoding", is_flag=True, help="Force decoding the audio file to get the duration")
@click.option("-t", "--show-file-type", is_flag=True, help="Show detected file-type")
@click.option("-r", "--show-sample-rate", is_flag=True, help="Show sample-rate")
@click.option("-c", "--show-channels", is_flag=True, help="Show number of channels")
@click.option("-s", "--show-samples", is_flag=True, help="Show number of samples (N/A if unavailable)")
@click.option(
    "-d", "--show-duration-hms", is_flag=True, help="Show duration in hours, minutes and seconds (N/A if unavailable)"
)
@click.option("-D", "--show-duration-seconds", is_flag=True, help="Show duration in seconds (N/A if unavailable)")
@click.option(
    "-b", "--show-bits-per-sample", is_flag=True, help="Show number of bits per sample (N/A if not applicable)"
)
@click.option(
    "-B", "--show-bitrate", is_flag=True, help="Show the bitrate averaged over the whole file (N/A if unavailable)"
)
@click.option("-p", "--show-precision", is_flag=True, help="Show estimated sample precision in bits")
@click.option("-e", "--show-encoding", is_flag=True, help="Show the name of the audio encoding")
@click.option("-a", "--show-comments", is_flag=True, help="Show file comments (annotations) if available")
def main(
    audio_files: Any,
    stream_id: int = 0,
    force_decoding: bool = False,
    show_file_type: bool = False,
    show_sample_rate: bool = False,
    show_channels: bool = False,
    show_samples: bool = False,
    show_duration_hms: bool = False,
    show_duration_seconds: bool = False,
    show_bits_per_sample: bool = False,
    show_bitrate: bool = False,
    show_precision: bool = False,
    show_encoding: bool = False,
    show_comments: bool = False,
):
    """
    Print the information of audio files.

    Args:
        audio_files: The audio files, paths to audio files, or stdin.
        stream_id: The index of the stream to load.
        force_decoding: Whether to force decoding the audio file to get the duration.
        show_file_type: Show detected file-type.
        show_sample_rate: Show sample-rate.
        show_channels: Show number of channels.
        show_samples: Show number of samples (0 if unavailable).
        show_duration_hms: Show duration in hours, minutes and seconds (0 if unavailable).
        show_duration_seconds: Show duration in seconds (0 if unavailable).
        show_bits_per_sample: Show number of bits per sample (0 if not applicable).
        show_bitrate: Show the bitrate averaged over the whole file (0 if unavailable).
        show_precision: Show estimated sample precision in bits.
        show_encoding: Show the name of the audio encoding.
        show_comments: Show file comments (annotations) if available.
    """
    # If no files are provided, use stdin
    if not audio_files:
        # Create a file object for stdin
        stdin_file = click.File(mode="rb").convert("-", None, None)
        audio_files = [stdin_file]

    # Initialize total duration and show_any flag
    total_duration = 0.0
    show_any = any(
        [
            show_file_type,
            show_sample_rate,
            show_channels,
            show_samples,
            show_duration_hms,
            show_duration_seconds,
            show_bits_per_sample,
            show_bitrate,
            show_precision,
            show_encoding,
            show_comments,
        ]
    )

    # Process each audio file
    for audio_file in audio_files:
        info = audiolab.info(audio_file, stream_id, force_decoding)
        # If no specific options are selected, show all information (default behavior)
        if not show_any:
            print(info)
            # Accumulate total duration
            total_duration += info.duration or 0.0

        # Display information based on selected options
        if show_file_type:
            print(info.container.format.name)
        if show_sample_rate:
            print(info.sample_rate)
        if show_channels:
            print(info.channels)
        if show_samples:
            print(info.num_samples or 0)
        if show_duration_hms:
            print(Info.format_duration(info.duration))
        if show_duration_seconds:
            print(info.duration or 0)
        if show_bits_per_sample:
            print(info.precision)
        if show_bitrate:
            print(Info.format_bit_rate(info.bit_rate))
        if show_precision:
            print(info.precision)
        if show_encoding:
            print(info.codec.long_name)
        if show_comments:
            if info.metadata:
                for key, value in info.metadata.items():
                    print(f"{key}: {value}")

    # Print total duration if any files were processed and any options were selected
    if len(audio_files) > 1 and not show_any:
        print(f"\nTotal duration of {len(audio_files)} files: {Info.format_duration(total_duration)}")



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from base64 import b64encode
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Tuple, Union

import av
import numpy as np

from audiolab.av import aformat, clip, from_ndarray, get_format, get_format_dtype, split_audio_frame, to_ndarray
from audiolab.av.typing import AudioFormat, ContainerFormat, Dtype
from audiolab.reader import Reader, StreamReader, info, load_audio
from audiolab.writer import Writer, save_audio

if TYPE_CHECKING:
    from lhotse import Recording
    from lhotse.cut import Cut


def encode(
    audio: Union[str, Path, np.ndarray, Cut, Recording],
    rate: Optional[int] = None,
    dtype: Optional[Dtype] = None,
    is_planar: bool = False,
    format: Optional[AudioFormat] = None,
    to_mono: bool = False,
    make_wav: bool = True,
    container_format: ContainerFormat = "wav",
) -> Tuple[str, int]:
    """
    Transform an audio to a PCM bytestring.

    Args:
        audio: The audio file, ndarray, Cut, Recording, etc.
        rate: The sample rate of the audio.
        dtype: The data type of the audio.
        is_planar: Whether the audio is planar.
        format: The format of the audio.
        to_mono: Whether to convert the audio to mono.
        make_wav: Whether to make the audio a WAV file.
        container_format: The format of the audio container.
    Returns:
        The audio as a PCM bytestring and the sample rate of the audio.
    """
    if isinstance(audio, (str, Path)):
        audio, rate = load_audio(audio, dtype=dtype, is_planar=is_planar, format=format, rate=rate)
    elif not isinstance(audio, np.ndarray):
        if rate is not None:
            audio = audio.resample(rate)
        rate = audio.sampling_rate
        audio = audio.load_audio()

    audio = clip(audio, np.int16)
    if to_mono and audio.ndim == 2 and audio.shape[0] > 1:
        audio = audio.mean(axis=0, keepdims=True).astype(audio.dtype)
    if make_wav:
        assert rate is not None
        bytestream = BytesIO()
        if isinstance(container_format, av.ContainerFormat):
            container_format = container_format.name
        save_audio(bytestream, audio, rate, container_format=container_format)
        audio = b64encode(bytestream.getvalue()).decode("ascii")
        audio = f"data:audio/{container_format};base64,{audio}"
    else:
        audio = np.ascontiguousarray(audio)
        audio = b64encode(audio).decode("ascii")
    return audio, rate


__all__ = [
    "Reader",
    "StreamReader",
    "Writer",
    "aformat",
    "encode",
    "from_ndarray",
    "get_format",
    "get_format_dtype",
    "info",
    "load_audio",
    "save_audio",
    "split_audio_frame",
    "to_ndarray",
]



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from fractions import Fraction
from typing import Optional, Tuple

import av
import numpy as np

from audiolab.av.format import get_format_dtype
from audiolab.av.typing import AudioFormat, AudioLayout, Seconds

logger = logging.getLogger(__name__)


def clip(ndarray: np.ndarray, dtype: np.dtype) -> np.ndarray:
    """
    Clip the ndarray to the given data type.

    Args:
        ndarray: The ndarray to clip.
        dtype: The data type to clip the ndarray to.
    """
    ndarray = ndarray.numpy() if "tensor" in ndarray.__class__.__name__.lower() else ndarray
    source_type = ndarray.dtype
    target_type = np.dtype(dtype)
    assert source_type.kind in ("f", "i", "u") and target_type.kind in ("f", "i", "u")

    min_value = np.iinfo(source_type).min if source_type.kind in ("i", "u") else -1
    max_value = np.iinfo(source_type).max if source_type.kind in ("i", "u") else 1
    if np.any(ndarray < min_value) or np.any(ndarray > max_value):
        logger.warning(f"{source_type} array out of range: {ndarray.min()} ~ {ndarray.max()}")
        ndarray = np.clip(ndarray, min_value, max_value)

    ndarray = ndarray.astype(np.float64)
    if source_type.kind == "u":
        ndarray = ndarray / np.iinfo(source_type).max * 2 - 1
    elif source_type.kind == "i":
        ndarray = ndarray / np.iinfo(source_type).max

    if target_type.kind == "u":
        ndarray = ((ndarray + 1) * 0.5 * np.iinfo(target_type).max).round()
    elif target_type.kind == "i":
        ndarray = (ndarray * np.iinfo(target_type).max).round()
    return ndarray.astype(dtype)


def from_ndarray(
    ndarray: np.ndarray,
    format: AudioFormat,
    layout: AudioLayout,
    rate: int,
    pts: Optional[int] = None,
    time_base: Optional[Fraction] = None,
) -> av.AudioFrame:
    """
    Create an AudioFrame from an ndarray.

    Args:
        ndarray: The ndarray to create the AudioFrame from.
        format: The format of the AudioFrame.
        layout: The layout of the AudioFrame.
        rate: The sample rate of the AudioFrame.
        pts: The presentation timestamp of the AudioFrame.
        time_base: The time base of the AudioFrame.
    Returns:
        The AudioFrame.
    """
    if isinstance(format, str):
        format = av.AudioFormat(format)
    if format.is_packed:
        # [num_channels, num_samples] => [1, num_channels * num_samples]
        ndarray = ndarray.T.reshape(1, -1)
    if isinstance(layout, str):
        layout = av.AudioLayout(layout)

    dtype = get_format_dtype(format)
    ndarray = clip(ndarray, dtype)
    frame = av.AudioFrame.from_ndarray(ndarray, format=format.name, layout=layout)
    frame.rate = rate
    if pts is not None:
        frame.pts = pts
    if time_base is not None:
        frame.time_base = time_base
    return frame


def to_ndarray(frame: av.AudioFrame) -> np.ndarray:
    """
    Convert an AudioFrame to an ndarray.

    Args:
        frame: The AudioFrame to convert.
            * shape of packed frame: [num_channels, num_samples]
            * shape of planar frame: [1, num_channels * num_samples]
    Returns:
        The ndarray.
            * shape: [num_channels, num_samples]
    """
    ndarray = frame.to_ndarray()
    if frame.format.is_packed:
        ndarray = ndarray.reshape(-1, frame.layout.nb_channels).T
    return ndarray


def split_audio_frame(frame: av.AudioFrame, offset: Seconds) -> Tuple[av.AudioFrame, av.AudioFrame]:
    """
    Split an AudioFrame into two AudioFrames.

    Args:
        frame: The AudioFrame to split.
        offset: The offset to split the AudioFrame at.
    Returns:
        The two AudioFrames.
    """
    offset = int(offset * frame.rate)
    if offset <= 0:
        return None, frame
    # number of audio samples (per channel)
    if offset > frame.samples:
        return frame, None

    ndarray = to_ndarray(frame)
    left, right = ndarray[:, :offset], ndarray[:, offset:]
    if frame.format.is_packed:
        left, right = left.T.reshape(1, -1), right.T.reshape(1, -1)
    left = av.AudioFrame.from_ndarray(left, format=frame.format.name, layout=frame.layout)
    right = av.AudioFrame.from_ndarray(right, format=frame.format.name, layout=frame.layout)
    left.rate, right.rate = frame.rate, frame.rate
    if frame.pts is not None:
        left.pts, right.pts = frame.pts, frame.pts + offset
    if frame.time_base is not None:
        left.time_base, right.time_base = frame.time_base, frame.time_base
    return left, right



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from typing import Iterator, List, Optional

import av

from audiolab.av import AudioGraph, aformat
from audiolab.av.typing import AudioFormat, Dtype, Filter
from audiolab.av.utils import is_streamable


class StreamReader:
    def __init__(
        self,
        filters: Optional[List[Filter]] = None,
        dtype: Optional[Dtype] = None,
        is_planar: bool = False,
        format: Optional[AudioFormat] = None,
        rate: Optional[int] = None,
        to_mono: bool = False,
        frame_size: Optional[int] = 1024,
        return_ndarray: bool = True,
    ):
        """
        Create a StreamReader object.

        Args:
            filters: The filters to apply to the audio stream.
            dtype: The data type of the output audio frames.
            is_planar: Whether the output audio frames are planar.
            format: The format of the output audio frames.
            rate: The sample rate of the output audio frames.
            to_mono: Whether to convert the output audio frames to mono.
            frame_size: The frame size of the audio frames.
            return_ndarray: Whether to return the audio frames as ndarrays.
        Returns:
            The StreamReader object.
        """
        self._codec_context = None
        self._graph = None
        self.bytestream = BytesIO()
        self.bytes_per_decode_attempt = 0
        if not all([dtype is None, format is None, rate is None, to_mono is None]):
            filters = filters or []
            filters.append(aformat(dtype, is_planar, format, rate, to_mono))
        self.filters = filters
        self.frame_size = frame_size
        self.offset = None
        self.return_ndarray = return_ndarray
        self.packet = None

    @property
    def codec_context(self) -> Optional[av.CodecContext]:
        """
        Get the codec context of the audio stream.

        Returns:
            The codec context of the audio stream.
        """
        if self._codec_context is None:
            if self.packet is None:
                return None
            self._codec_context = self.packet.stream.codec_context
            assert is_streamable(self._codec_context), "Only support streamable codec"
        return self._codec_context

    @property
    def graph(self) -> Optional[AudioGraph]:
        """
        Get the audio graph of the audio stream.

        Returns:
            The audio graph of the audio stream.
        """
        if self._graph is None:
            if self.packet is None:
                return None
            self._graph = AudioGraph(
                stream=self.packet.stream,
                filters=self.filters,
                frame_size=self.frame_size,
                return_ndarray=self.return_ndarray,
            )
        return self._graph

    @property
    def is_decoded(self) -> bool:
        """
        Check if the current frame is decoded.

        Returns:
            Whether the current frame is decoded.
        """
        # o: current frame
        # pts: self.offset, frame.pts, packet.pts
        # +---+---+---+---+---+
        # | x | x | x | o |   |
        # +---+---+---+---+---+
        #             ↑
        #             pts
        if self.offset is None:
            return False
        if self.packet.pts is None:
            return False
        if self.offset <= self.packet.pts:
            return False
        return True

    def should_decode(self, partial: bool = False) -> bool:
        """
        Check if the current frame should be decoded.

        Args:
            partial: Whether to decode a partial frame.
        Returns:
            Whether the current frame should be decoded.
        """
        if partial:
            return True
        if self.bytes_per_decode_attempt * 2 < self.frame_size:
            return False
        self.bytes_per_decode_attempt = 0
        return True

    def ready_for_decoding(self, partial: bool = False) -> bool:
        """
        Check if the current frame is ready for decoding.

        Args:
            partial: Whether to decode a partial frame.
        Returns:
            Whether the current frame is ready for decoding.
        """
        if self.packet.pts is None and not partial:
            return False
        return not self.is_decoded

    def push(self, chunk: bytes):
        """
        Push a chunk of audio data to the audio stream.

        Args:
            chunk: The chunk of audio data to push.
        """
        self.bytestream.write(chunk)
        self.bytes_per_decode_attempt += len(chunk)

    def pull(self, partial: bool = False) -> Optional[Iterator[av.AudioFrame]]:
        """
        Pull an audio frame from the audio stream.

        Args:
            partial: Whether to pull a partial frame.
        Yields:
            The audio frame.
        """
        if not self.should_decode(partial):
            return
        try:
            self.bytestream.seek(0)
            container = av.open(self.bytestream)
            for packet in container.demux():
                self.packet = packet
                if not self.ready_for_decoding(partial):
                    continue
                for frame in self.codec_context.decode(packet):
                    self.offset = frame.pts + int(frame.samples / packet.stream.rate / packet.stream.time_base)
                    self.graph.push(frame)
                    yield from self.graph.pull()
                yield from self.graph.pull(partial=partial)
        except (av.EOFError, av.InvalidDataError, av.OSError, av.PermissionError):
            pass

    def reset(self):
        """
        Reset the StreamReader object.

        Returns:
            The StreamReader object.
        """
        self._codec_context = None
        self._graph = None
        self.bytestream = BytesIO()
        self.bytes_per_decode_attempt = 0
        self.offset = None
        self.packet = None



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import av

from audiolab.av.typing import AudioLayoutEnum

"""
$ ffmpeg -layouts
"""
standard_channel_layouts = (
    "mono",
    "stereo",
    "2.1",
    "3.0",
    "3.0(back)",
    "4.0",
    "quad",
    "quad(side)",
    "3.1",
    "5.0",
    "5.0(side)",
    "4.1",
    "5.1",
    "5.1(side)",
    "6.0",
    "6.0(front)",
    "3.1.2",
    "hexagonal",
    "6.1",
    "6.1(back)",
    "6.1(front)",
    "7.0",
    "7.0(front)",
    "7.1",
    "7.1(wide)",
    "7.1(wide-side)",
    "5.1.2",
    "octagonal",
    "cube",
    "5.1.4",
    "7.1.2",
    "7.1.4",
    "7.2.3",
    "9.1.4",
    "hexadecagonal",
    "downmix",
    "22.2",
)
audio_layouts: Dict[str, av.AudioLayout] = {name: av.AudioLayout(name) for name in standard_channel_layouts}
AudioLayout = AudioLayoutEnum("AudioLayout", audio_layouts)



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

import av

from audiolab.av.format import get_format_dtype
from audiolab.av.frame import to_ndarray
from audiolab.av.typing import AudioFormat, AudioFrame, AudioLayout, Codec, ContainerFormat, Dtype
from audiolab.writer.writer import Writer


def save_audio(
    file: Any,
    frame: AudioFrame,
    rate: int,
    codec: Optional[Codec] = None,
    channels: Optional[int] = None,
    dtype: Optional[Dtype] = None,
    is_planar: Optional[bool] = None,
    format: Optional[AudioFormat] = None,
    layout: Optional[AudioLayout] = None,
    container_format: Optional[ContainerFormat] = None,
    options: Optional[Dict[str, str]] = None,
    **kwargs,
):
    """
    Save an audio frame to a file.

    Args:
        file: The audio file, path to audio file, bytes, etc.
        rate: The sample rate of the audio stream.
        codec: The codec of the audio container.
        channels: The number of channels of the audio stream.
        dtype: The data type of the audio stream.
        is_planar: Whether the audio stream is planar.
        format: The format of the audio stream.
        layout: The layout of the audio stream.
        container_format: The format of the audio container.
        options: The options of the audio stream.
        **kwargs: Additional arguments for the Writer class.
    """
    if isinstance(frame, av.AudioFrame):
        if format is None:
            dtype = dtype or get_format_dtype(frame.format)
            is_planar = is_planar or frame.format.is_planar
        frame = to_ndarray(frame)

    channels = 1 if frame.ndim == 1 else frame.shape[0]
    assert channels in (1, 2)
    writer = Writer(file, rate, codec, channels, dtype, is_planar, format, layout, container_format, options, **kwargs)
    writer.write(frame)
    writer.close()


__all__ = ["Writer", "save_audio"]



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
from fractions import Fraction
from typing import List, Optional

import av
import numpy as np
from av.filter import Graph

from audiolab.av.format import get_format
from audiolab.av.frame import from_ndarray, to_ndarray
from audiolab.av.typing import AudioFormat, AudioFrame, AudioLayout, Dtype, Filter


class AudioGraph:
    def __init__(
        self,
        stream: Optional[av.AudioStream] = None,
        rate: Optional[int] = None,
        dtype: Optional[Dtype] = None,
        is_planar: bool = False,
        format: Optional[AudioFormat] = None,
        layout: Optional[AudioLayout] = None,
        channels: Optional[int] = None,
        name: Optional[str] = None,
        time_base: Optional[Fraction] = None,
        filters: Optional[List[Filter]] = None,
        frame_size: int = -1,
        return_ndarray: bool = True,
    ):
        """
        Create an AudioGraph.

        Args:
            stream: The stream to create the AudioGraph from.
            rate: The sample rate of the input audio buffer.
            dtype: The data type of the input audio buffer.
            is_planar: Whether the input audio buffer is planar.
            format: The format of the input audio buffer.
            layout: The layout of the input audio buffer.
            channels: The number of channels of the input audio buffer.
            name: The name of the input audio buffer.
            time_base: The time base of the input audio buffer.
            filters: The filters to apply to the input audio buffer.
            frame_size: The frame size of the output audio buffer.
            return_ndarray: Whether to return the output audio frames as ndarrays.
        """
        self.filters = filters or []
        self.graph = Graph()
        if stream is None:
            if format is None:
                format = get_format(dtype, is_planar)
            abuffer = self.graph.add_abuffer(
                sample_rate=rate, format=format, layout=layout, channels=channels, name=name, time_base=time_base
            )
            self.rate = rate
            self.format = format.name if isinstance(format, av.AudioFormat) else format
            self.layout = layout
        else:
            abuffer = self.graph.add_abuffer(template=stream)
            self.rate = stream.rate
            self.format = stream.format.name
            self.layout = stream.layout
        nodes = [abuffer]
        for _filter in self.filters:
            name, args, kwargs = (
                (_filter, None, {}) if isinstance(_filter, str) else ((*_filter, {}) if len(_filter) == 2 else _filter)
            )
            nodes.append(self.graph.add(name, args, **kwargs))
        nodes.append(self.graph.add("abuffersink"))
        self.graph.link_nodes(*nodes).configure()

        if frame_size > 0:
            self.graph.set_audio_frame_size(frame_size)
        self.return_ndarray = return_ndarray

    def push(self, frame: AudioFrame):
        """
        Push an audio frame to the graph.

        Args:
            frame: The audio frame to push.
                * shape of ndarray: [num_channels, num_samples]
        """
        if isinstance(frame, np.ndarray):
            frame = from_ndarray(frame, self.format, self.layout, self.rate)
        self.graph.push(frame)

    def pull(self, partial: bool = False, return_ndarray: Optional[bool] = None):
        """
        Pull an audio frame from the graph.

        Args:
            partial: Whether to pull a partial frame.
            return_ndarray: Whether to return the audio frame as an ndarray.
                * shape of ndarray: [num_channels, num_samples]
        """
        if partial:
            self.graph.push(None)
        while True:
            try:
                frame = self.graph.pull()
                if return_ndarray is None:
                    return_ndarray = self.return_ndarray
                if return_ndarray:
                    yield to_ndarray(frame), frame.rate
                else:
                    yield frame
            except av.EOFError:
                break
            except av.FFmpegError as e:
                if e.errno != errno.EAGAIN:
                    raise
                break
